"""Protocol and CAN constants"""

ADVERTISEMENT_TIME_EEPROM_TO_MS = 0.625
"""Stores the conversion rate for the EEPROM advertisement times

See also:

- https://mytoolit.github.io/Documentation/#page-system-configuration

"""

SENSOR_NODE_NUMBER_SELF_ADDRESSING = 0xFF
"""Sensor Node Number for self addressing mode"""
